import autogalaxy as ag
import autogalaxy as ag
import autogalaxy.plot as aplt

grid = ag.Grid.uniform(shape_2d=(101, 101), pixel_scales=0.05)

mass_profile = ag.mp.EllipticalIsothermal(
    einstein_radius=1.0, elliptical_comps=(0.0, 0.05263)
)

aplt.MassProfile.figure_convergence(mass_profile=mass_profile, grid=grid)



import autogalaxy as ag
import numpy as np
import pytest
from autogalaxy import exc


grid = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [2.0, 4.0]])


class TestMassSheet:
    def test__convergence__correct_values(self):

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        convergence = mass_sheet.convergence_from_grid(grid=np.array([[1.0, 0.0]]))

        assert convergence[0] == pytest.approx(1.0, 1e-3)

        convergence = mass_sheet.convergence_from_grid(
            grid=np.array([[1.0, 0.0], [3.0, 3.0], [5.0, -9.0]])
        )

        assert convergence[0] == pytest.approx(1.0, 1e-3)
        assert convergence[1] == pytest.approx(1.0, 1e-3)
        assert convergence[2] == pytest.approx(1.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=-3.0)

        convergence = mass_sheet.convergence_from_grid(
            grid=np.array([[1.0, 0.0], [3.0, 3.0], [5.0, -9.0]])
        )

        assert convergence[0] == pytest.approx(-3.0, 1e-3)
        assert convergence[1] == pytest.approx(-3.0, 1e-3)
        assert convergence[2] == pytest.approx(-3.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        convergence = mass_sheet.convergence_from_grid(
            grid=ag.Grid.manual_2d(
                [[[1.0, 0.0], [1.0, 0.0]], [[1.0, 0.0], [1.0, 0.0]]],
                sub_size=2,
                pixel_scales=(1.0, 1.0),
            )
        )

        assert convergence[0] == pytest.approx(1.0, 1e-3)
        assert convergence[1] == pytest.approx(1.0, 1e-3)
        assert convergence[2] == pytest.approx(1.0, 1e-3)
        assert convergence[3] == pytest.approx(1.0, 1e-3)

        convergence = mass_sheet.convergence_from_grid(grid=np.array([[1.0, 0.0]]))

        assert convergence[0] == pytest.approx(1.0, 1e-3)

    def test__potential__correct_values(self):
        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        potential = mass_sheet.potential_from_grid(
            grid=np.array([[1.0, 0.0], [3.0, 3.0], [5.0, -9.0]])
        )

        assert potential[0] == pytest.approx(0.0, 1e-3)
        assert potential[1] == pytest.approx(0.0, 1e-3)
        assert potential[2] == pytest.approx(0.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        potential = mass_sheet.potential_from_grid(
            grid=ag.Grid.manual_2d(
                [[[1.0, 0.0], [1.0, 0.0]], [[1.0, 0.0], [1.0, 0.0]]],
                sub_size=2,
                pixel_scales=(1.0, 1.0),
            )
        )

        assert potential[0] == pytest.approx(0.0, 1e-3)
        assert potential[1] == pytest.approx(0.0, 1e-3)
        assert potential[2] == pytest.approx(0.0, 1e-3)
        assert potential[3] == pytest.approx(0.0, 1e-3)

        potential = mass_sheet.potential_from_grid(grid=np.array([[1.0, 0.0]]))

        assert potential[0] == pytest.approx(0.0, 1e-3)

    def test__deflections__correct_values(self):
        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[1.0, 0.0]]))

        assert deflections[0, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.0, 1e-3)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[2.0, 0.0]]))

        assert deflections[0, 0] == pytest.approx(2.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=-1.0)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[1.0, 0.0]]))

        assert deflections[0, 0] == pytest.approx(-1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.0, 1e-3)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[2.0, 0.0]]))

        assert deflections[0, 0] == pytest.approx(-2.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=2.0)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[2.0, 0.0]]))

        assert deflections[0, 0] == pytest.approx(4.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        # The radial coordinate at (1.0, 1.0) is sqrt(2)
        # This is decomposed into (y,x) angles of sin(45) = cos(45) = sqrt(2) / 2.0
        # Thus, for a mass sheet, the deflection angle is (sqrt(2) * sqrt(2) / 2.0) = 1.0

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[1.0, 1.0]]))
        assert deflections[0, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(1.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=2.0)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[1.0, 1.0]]))
        assert deflections[0, 0] == pytest.approx(2.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(2.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=2.0)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[2.0, 2.0]]))
        assert deflections[0, 0] == pytest.approx(4.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(4.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        # The radial coordinate at (2.0, 1.0) is sqrt(5)
        # This gives an angle of 26.5650512 degrees between the 1.0 and np.sqrt(5) of the triangle
        # This is decomposed into y angle of cos(26.5650512 degrees) = 0.8944271
        # This is decomposed into x angle of sin(26.5650512 degrees) = 0.4472135
        # Thus, for a mass sheet, the deflection angles are:
        deflections = mass_sheet.deflections_from_grid(grid=np.array([[2.0, 1.0]]))
        assert deflections[0, 0] == pytest.approx(0.8944271 * np.sqrt(5), 1e-3)
        assert deflections[0, 1] == pytest.approx(0.4472135 * np.sqrt(5), 1e-3)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[-1.0, -1.0]]))
        assert deflections[0, 0] == pytest.approx(-1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(-1.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(1.0, 2.0), kappa=1.0)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[2.0, 3.0]]))
        assert deflections[0, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(1.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(1.0, 2.0), kappa=-1.0)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[2.0, 3.0]]))
        assert deflections[0, 0] == pytest.approx(-1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(-1.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        deflections = mass_sheet.deflections_from_grid(
            grid=ag.Grid.manual_2d(
                [[[1.0, 0.0], [1.0, 0.0]], [[1.0, 0.0], [1.0, 0.0]]],
                sub_size=2,
                pixel_scales=(1.0, 1.0),
            )
        )

        assert deflections[0, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[1, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[2, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[3, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.0, 1e-3)
        assert deflections[1, 1] == pytest.approx(0.0, 1e-3)
        assert deflections[2, 1] == pytest.approx(0.0, 1e-3)
        assert deflections[3, 1] == pytest.approx(0.0, 1e-3)

        deflections = mass_sheet.deflections_from_grid(grid=np.array([[1.0, 0.0]]))

        assert deflections[0, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.0, 1e-3)

    def test__deflections__change_geometry(self):
        mass_sheet_0 = ag.mp.MassSheet(centre=(0.0, 0.0))
        mass_sheet_1 = ag.mp.MassSheet(centre=(1.0, 1.0))
        deflections_0 = mass_sheet_0.deflections_from_grid(grid=np.array([[1.0, 1.0]]))
        deflections_1 = mass_sheet_1.deflections_from_grid(grid=np.array([[0.0, 0.0]]))
        assert deflections_0[0, 0] == pytest.approx(-deflections_1[0, 0], 1e-5)
        assert deflections_0[0, 1] == pytest.approx(-deflections_1[0, 1], 1e-5)

        mass_sheet_0 = ag.mp.MassSheet(centre=(0.0, 0.0))
        mass_sheet_1 = ag.mp.MassSheet(centre=(0.0, 0.0))
        deflections_0 = mass_sheet_0.deflections_from_grid(grid=np.array([[1.0, 0.0]]))
        deflections_1 = mass_sheet_1.deflections_from_grid(grid=np.array([[0.0, 1.0]]))
        assert deflections_0[0, 0] == pytest.approx(deflections_1[0, 1], 1e-5)
        assert deflections_0[0, 1] == pytest.approx(deflections_1[0, 0], 1e-5)

    def test__multiple_coordinates_in__multiple_coordinates_out(self):
        mass_sheet = ag.mp.MassSheet(centre=(1.0, 2.0), kappa=1.0)

        deflections = mass_sheet.deflections_from_grid(
            grid=np.array([[2.0, 3.0], [2.0, 3.0], [2.0, 3.0]])
        )

        assert deflections[0, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(1.0, 1e-3)
        assert deflections[1, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[1, 1] == pytest.approx(1.0, 1e-3)
        assert deflections[2, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[2, 1] == pytest.approx(1.0, 1e-3)

        mass_sheet = ag.mp.MassSheet(centre=(0.0, 0.0), kappa=1.0)

        deflections = mass_sheet.deflections_from_grid(
            grid=np.array([[1.0, 1.0], [2.0, 2.0], [1.0, 1.0], [2.0, 2.0]])
        )

        assert deflections[0, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[0, 1] == pytest.approx(1.0, 1e-3)

        assert deflections[1, 0] == pytest.approx(2.0, 1e-3)
        assert deflections[1, 1] == pytest.approx(2.0, 1e-3)

        assert deflections[2, 0] == pytest.approx(1.0, 1e-3)
        assert deflections[2, 1] == pytest.approx(1.0, 1e-3)

        assert deflections[3, 0] == pytest.approx(2.0, 1e-3)
        assert deflections[3, 1] == pytest.approx(2.0, 1e-3)

    def test__outputs_are_autoarrays(self):

        grid = ag.Grid.uniform(shape_2d=(2, 2), pixel_scales=1.0, sub_size=1)

        mass_sheet = ag.mp.MassSheet()

        convergence = mass_sheet.convergence_from_grid(grid=grid)

        assert convergence.shape_2d == (2, 2)

        potential = mass_sheet.potential_from_grid(grid=grid)

        assert potential.shape_2d == (2, 2)

        deflections = mass_sheet.deflections_from_grid(grid=grid)

        assert deflections.shape_2d == (2, 2)


class TestExternalShear:
    def test__convergence_returns_zeros(self):

        shear = ag.mp.ExternalShear(elliptical_comps=(0.1, 0.0))
        convergence = shear.convergence_from_grid(grid=np.array([[0.1, 0.1]]))
        assert (convergence == np.array([0.0])).all()

        shear = ag.mp.ExternalShear(elliptical_comps=(0.1, 0.0))
        convergence = shear.convergence_from_grid(
            grid=np.array([[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]])
        )
        assert (convergence == np.array([0.0, 0.0, 0.0])).all()

        convergence = shear.convergence_from_grid(
            grid=ag.Grid.manual_2d(
                [[[1.0, 0.0], [1.0, 0.0]], [[1.0, 0.0], [1.0, 0.0]]],
                sub_size=2,
                pixel_scales=(1.0, 1.0),
            )
        )

        assert convergence[0] == pytest.approx(0.0, 1e-3)
        assert convergence[1] == pytest.approx(0.0, 1e-3)
        assert convergence[2] == pytest.approx(0.0, 1e-3)
        assert convergence[3] == pytest.approx(0.0, 1e-3)

        convergence = shear.convergence_from_grid(grid=np.array([[1.0, 0.0]]))

        assert convergence[0] == pytest.approx(0.0, 1e-3)

    def test__potential_returns_zeros(self):
        shear = ag.mp.ExternalShear(elliptical_comps=(0.1, 0.0))
        potential = shear.potential_from_grid(grid=np.array([[0.1, 0.1]]))
        assert (potential == np.array([[0.0, 0.0]])).all()

        shear = ag.mp.ExternalShear(elliptical_comps=(0.1, 0.0))
        potential = shear.potential_from_grid(
            grid=np.array([[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]])
        )
        assert (potential == np.array([0.0, 0.0, 0.0])).all()

        potential = shear.potential_from_grid(
            grid=ag.Grid.manual_2d(
                [[[1.0, 0.0], [1.0, 0.0]], [[1.0, 0.0], [1.0, 0.0]]],
                sub_size=2,
                pixel_scales=(1.0, 1.0),
            )
        )

        assert potential[0] == pytest.approx(0.0, 1e-3)
        assert potential[1] == pytest.approx(0.0, 1e-3)
        assert potential[2] == pytest.approx(0.0, 1e-3)
        assert potential[3] == pytest.approx(0.0, 1e-3)

        potential = shear.potential_from_grid(grid=np.array([[1.0, 0.0]]))

        assert potential[0] == pytest.approx(0.0, 1e-3)

    def test__deflections_correct_values(self):

        shear = ag.mp.ExternalShear(elliptical_comps=(0.1, 0.0))
        deflections = shear.deflections_from_grid(grid=np.array([[0.1625, 0.1625]]))
        assert deflections[0, 0] == pytest.approx(0.01625, 1e-3)
        assert deflections[0, 1] == pytest.approx(0.01625, 1e-3)

        shear = ag.mp.ExternalShear(elliptical_comps=(0.1, -0.17320))
        deflections = shear.deflections_from_grid(grid=np.array([[0.1625, 0.1625]]))
        assert deflections[0, 0] == pytest.approx(0.04439, 1e-3)
        assert deflections[0, 1] == pytest.approx(-0.011895, 1e-3)

        deflections = shear.deflections_from_grid(
            grid=ag.Grid.manual_2d(
                [
                    [[0.1625, 0.1625], [0.1625, 0.1625]],
                    [[0.1625, 0.1625], [0.1625, 0.1625]],
                ],
                sub_size=2,
                pixel_scales=(1.0, 1.0),
            )
        )

        assert deflections[0, 0] == pytest.approx(0.04439, 1e-3)
        assert deflections[1, 0] == pytest.approx(0.04439, 1e-3)
        assert deflections[2, 0] == pytest.approx(0.04439, 1e-3)
        assert deflections[3, 0] == pytest.approx(0.04439, 1e-3)
        assert deflections[0, 1] == pytest.approx(-0.011895, 1e-3)
        assert deflections[1, 1] == pytest.approx(-0.011895, 1e-3)
        assert deflections[2, 1] == pytest.approx(-0.011895, 1e-3)
        assert deflections[3, 1] == pytest.approx(-0.011895, 1e-3)

        deflections = shear.deflections_from_grid(grid=np.array([[0.1625, 0.1625]]))

        assert deflections[0, 0] == pytest.approx(0.04439, 1e-3)
        assert deflections[0, 1] == pytest.approx(-0.011895, 1e-3)

    def test__outputs_are_autoarrays(self):

        grid = ag.Grid.uniform(shape_2d=(2, 2), pixel_scales=1.0, sub_size=1)

        shear = ag.mp.ExternalShear()

        convergence = shear.convergence_from_grid(grid=grid)

        assert convergence.shape_2d == (2, 2)

        potential = shear.potential_from_grid(grid=grid)

        assert potential.shape_2d == (2, 2)

        deflections = shear.deflections_from_grid(grid=grid)

        assert deflections.shape_2d == (2, 2)


class TestInputDeflections:
    def test__deflections_from_grid__grid_coordinates_overlap_image_grid_of_deflections(
        self,
    ):

        deflections_y = ag.Array.manual_2d(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )
        deflections_x = ag.Array.manual_2d(
            [[9.0, 8.0, 7.0], [6.0, 5.0, 4.0], [3.0, 2.0, 1.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )

        image_plane_grid = ag.Grid.uniform(
            shape_2d=deflections_y.shape_2d, pixel_scales=deflections_y.pixel_scales
        )

        input_deflections = ag.mp.InputDeflections(
            deflections_y=deflections_y,
            deflections_x=deflections_x,
            image_plane_grid=image_plane_grid,
        )

        grid = ag.Grid.uniform(
            shape_2d=deflections_y.shape_2d, pixel_scales=deflections_y.pixel_scales
        )

        deflections = input_deflections.deflections_from_grid(grid=grid)

        assert deflections[:, 0] == pytest.approx(deflections_y, 1.0e-4)
        assert deflections[:, 1] == pytest.approx(deflections_x, 1.0e-4)

        grid = ag.Grid.manual_1d(
            grid=np.array([[0.1, 0.0], [0.0, 0.0], [-0.1, -0.1]]),
            shape_2d=deflections_y.shape_2d,
            pixel_scales=deflections_y.pixel_scales,
        )

        deflections = input_deflections.deflections_from_grid(grid=grid)

        assert deflections[:, 0] == pytest.approx([2.0, 5.0, 7.0], 1.0e-4)
        assert deflections[:, 1] == pytest.approx([8.0, 5.0, 3.0], 1.0e-4)

        # input_deflections = ag.mp.InputDeflections(
        #     deflections_y=deflections_y,
        #     deflections_x=deflections_x,
        #     image_plane_grid=image_plane_grid,
        #     normalization_scale=2.0,
        # )
        #
        # deflections = input_deflections.deflections_from_grid(grid=grid)
        #
        # assert deflections[:, 0] == pytest.approx([4.0, 10.0, 14.0], 1.0e-4)
        # assert deflections[:, 1] == pytest.approx([16.0, 10.0, 6.0], 1.0e-4)

    def test__deflections_from_grid__grid_coordinates_dont_overlap_image_grid_of_deflections__uses_interpolation(
        self,
    ):

        deflections_y = ag.Array.manual_2d(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )
        deflections_x = ag.Array.manual_2d(
            [[9.0, 8.0, 7.0], [6.0, 5.0, 4.0], [3.0, 2.0, 1.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )

        image_plane_grid = ag.Grid.uniform(
            shape_2d=deflections_y.shape_2d, pixel_scales=deflections_y.pixel_scales
        )

        input_deflections = ag.mp.InputDeflections(
            deflections_y=deflections_y,
            deflections_x=deflections_x,
            image_plane_grid=image_plane_grid,
        )

        grid = ag.Grid.manual_1d(
            grid=np.array([[0.05, 0.03], [0.02, 0.01], [-0.08, -0.04]]),
            shape_2d=deflections_y.shape_2d,
            pixel_scales=deflections_y.pixel_scales,
        )

        deflections = input_deflections.deflections_from_grid(grid=grid)

        assert deflections[:, 0] == pytest.approx([3.8, 4.5, 7.0], 1.0e-4)
        assert deflections[:, 1] == pytest.approx([6.2, 5.5, 3.0], 1.0e-4)

    def test__deflections_from_grid__preload_grid_deflections_used_if_preload_grid_input(
        self,
    ):

        deflections_y = ag.Array.manual_2d(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )
        deflections_x = ag.Array.manual_2d(
            [[9.0, 8.0, 7.0], [6.0, 5.0, 4.0], [3.0, 2.0, 1.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )

        image_plane_grid = ag.Grid.uniform(
            shape_2d=deflections_y.shape_2d, pixel_scales=deflections_y.pixel_scales
        )

        grid = ag.Grid.manual_1d(
            grid=np.array([[0.05, 0.03], [0.02, 0.01], [-0.08, -0.04]]),
            shape_2d=deflections_y.shape_2d,
            pixel_scales=deflections_y.pixel_scales,
        )

        input_deflections = ag.mp.InputDeflections(
            deflections_y=deflections_y,
            deflections_x=deflections_x,
            image_plane_grid=image_plane_grid,
            preload_grid=grid,
        )

        input_deflections.preload_deflections[0, 0] = 1.0

        deflections = input_deflections.deflections_from_grid(grid=grid)

        assert (deflections == input_deflections.preload_deflections).all()

        # input_deflections = ag.mp.InputDeflections(
        #     deflections_y=deflections_y,
        #     deflections_x=deflections_x,
        #     image_plane_grid=image_plane_grid,
        #     preload_grid=grid,
        #     normalization_scale=2.0,
        # )
        #
        # input_deflections.preload_deflections[0, 0] = 1.0
        #
        # deflections = input_deflections.deflections_from_grid(grid=grid)
        #
        # assert (deflections == 2.0 * input_deflections.preload_deflections).all()

    def test__deflections_from_grid__input_grid_extends_beyond_image_plane_grid__raises_exception(
        self,
    ):

        deflections_y = ag.Array.manual_2d(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )
        deflections_x = ag.Array.manual_2d(
            [[9.0, 8.0, 7.0], [6.0, 5.0, 4.0], [3.0, 2.0, 1.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )

        image_plane_grid = ag.Grid.uniform(
            shape_2d=deflections_y.shape_2d, pixel_scales=deflections_y.pixel_scales
        )

        input_deflections = ag.mp.InputDeflections(
            deflections_y=deflections_y,
            deflections_x=deflections_x,
            image_plane_grid=image_plane_grid,
        )

        grid = ag.Grid.manual_1d(
            grid=np.array([[0.0999, 0.0]]),
            shape_2d=deflections_y.shape_2d,
            pixel_scales=deflections_y.pixel_scales,
        )
        input_deflections.deflections_from_grid(grid=grid)

        grid = ag.Grid.manual_1d(
            grid=np.array([[0.0, 0.0999]]),
            shape_2d=deflections_y.shape_2d,
            pixel_scales=deflections_y.pixel_scales,
        )
        input_deflections.deflections_from_grid(grid=grid)

        with pytest.raises(exc.ProfileException):
            grid = ag.Grid.manual_1d(
                grid=np.array([[0.11, 0.0]]),
                shape_2d=deflections_y.shape_2d,
                pixel_scales=deflections_y.pixel_scales,
            )
            input_deflections.deflections_from_grid(grid=grid)

            with pytest.raises(exc.ProfileException):
                grid = ag.Grid.manual_1d(
                    grid=np.array([[0.0, 0.11]]),
                    shape_2d=deflections_y.shape_2d,
                    pixel_scales=deflections_y.pixel_scales,
                )
                input_deflections.deflections_from_grid(grid=grid)

    def test__convergence_from_grid_potential_from_grid(self):

        deflections_y = ag.Array.manual_2d(
            [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )
        deflections_x = ag.Array.manual_2d(
            [[9.0, 8.0, 7.0], [6.0, 5.0, 4.0], [3.0, 2.0, 1.0]],
            pixel_scales=0.1,
            origin=(0.0, 0.0),
        )

        image_plane_grid = ag.Grid.uniform(
            shape_2d=deflections_y.shape_2d, pixel_scales=deflections_y.pixel_scales
        )

        input_deflections = ag.mp.InputDeflections(
            deflections_y=deflections_y,
            deflections_x=deflections_x,
            image_plane_grid=image_plane_grid,
        )

        grid = ag.Grid.uniform(
            shape_2d=deflections_y.shape_2d, pixel_scales=deflections_y.pixel_scales
        )

        convergence = input_deflections.convergence_from_grid(grid=grid)

        convergence_via_jacobian = input_deflections.convergence_via_jacobian_from_grid(
            grid=grid
        )

        assert (convergence == convergence_via_jacobian).all()

        potential = input_deflections.potential_from_grid(grid=grid)

        assert (potential == np.zeros(shape=(9,))).all()



import typing

import numpy as np
import pytest

from pyquad import quad_grid
from skimage import measure

import autogalaxy as ag
from autoarray.structures import grids
from autogalaxy import lensing
from autogalaxy.profiles import geometry_profiles


class MockEllipticalIsothermal(
    geometry_profiles.EllipticalProfile, lensing.LensingObject
):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        einstein_radius: float = 1.0,
    ):
        """
        Abstract class for elliptical mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        """
        super(MockEllipticalIsothermal, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        self.einstein_radius = einstein_radius

    @property
    def einstein_radius_rescaled(self):
        """Rescale the einstein radius by slope and axis_ratio, to reduce its degeneracy with other mass-profiles
        parameters"""
        return (1.0 / (1 + self.axis_ratio)) * self.einstein_radius

    def convergence_func(self, grid_radius):
        return self.einstein_radius_rescaled * (grid_radius ** 2) ** (-0.5)

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid(self, grid):
        """ Calculate the projected convergence at a given set of arc-second gridded coordinates.

        The `grid_like_to_structure` decorator reshapes the ndarrays the convergence is outputted on. See \
        *ag.grid_like_to_structure* for a description of the output.

        Parameters
        ----------
        grid : ag.Grid
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """

        covnergence_grid = np.zeros(grid.shape[0])

        grid_eta = self.grid_to_elliptical_radii(grid)

        for i in range(grid.shape[0]):
            covnergence_grid[i] = self.convergence_func(grid_eta[i])

        return covnergence_grid

    @staticmethod
    def potential_func(u, y, x, axis_ratio):
        eta_u = np.sqrt((u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u)))))
        return (
            (eta_u / u)
            * (eta_u) ** -1.0
            * eta_u
            / ((1 - (1 - axis_ratio ** 2) * u) ** 0.5)
        )

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def potential_from_grid(self, grid):
        """
        Calculate the potential at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : ag.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        potential_grid = quad_grid(
            self.potential_func, 0.0, 1.0, grid, args=(self.axis_ratio)
        )[0]

        return self.einstein_radius_rescaled * self.axis_ratio * potential_grid

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        For coordinates (0.0, 0.0) the analytic calculation of the deflection angle gives a NaN. Therefore, \
        coordinates at (0.0, 0.0) are shifted slightly to (1.0e-8, 1.0e-8).

        Parameters
        ----------
        grid : ag.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """
        factor = (
            2.0
            * self.einstein_radius_rescaled
            * self.axis_ratio
            / np.sqrt(1 - self.axis_ratio ** 2)
        )

        psi = np.sqrt(
            np.add(
                np.multiply(self.axis_ratio ** 2, np.square(grid[:, 1])),
                np.square(grid[:, 0]),
            )
        )

        deflection_y = np.arctanh(
            np.divide(np.multiply(np.sqrt(1 - self.axis_ratio ** 2), grid[:, 0]), psi)
        )
        deflection_x = np.arctan(
            np.divide(np.multiply(np.sqrt(1 - self.axis_ratio ** 2), grid[:, 1]), psi)
        )
        return self.rotate_grid_from_profile(
            np.multiply(factor, np.vstack((deflection_y, deflection_x)).T)
        )

    @property
    def is_point_mass(self):
        return False

    @property
    def mass_profiles(self):
        return [self]

    @property
    def mass_profile_centres(self):
        return [self.centre]


class MockSphericalIsothermal(MockEllipticalIsothermal):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        einstein_radius: float = 1.0,
    ):
        """
        Abstract class for elliptical mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        """
        super(MockSphericalIsothermal, self).__init__(
            centre=centre, elliptical_comps=(0.0, 0.0), einstein_radius=einstein_radius
        )

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def potential_from_grid(self, grid):
        """
        Calculate the potential at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : ag.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """
        eta = self.grid_to_elliptical_radii(grid)
        return 2.0 * self.einstein_radius_rescaled * eta

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : ag.Grid
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """
        return self.grid_to_grid_cartesian(
            grid=grid,
            radius=np.full(grid.shape[0], 2.0 * self.einstein_radius_rescaled),
        )


class MockGalaxy(lensing.LensingObject):
    def __init__(self, mass_profiles):
        self._mass_profiles = mass_profiles

    @property
    def mass_profiles(self):
        return self._mass_profiles

    @property
    def mass_profile_centres(self):
        return [mass_profile.centre for mass_profile in self.mass_profiles]


class TestDeflectionsMagnitudes:
    def test__compare_sis_deflection_magnitudes_to_known_values(self):
        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=1.0)

        grid = ag.GridIrregularGrouped([(1.0, 0.0), (0.0, 1.0)])

        deflection_magnitudes = sis.deflection_magnitudes_from_grid(grid=grid)

        assert deflection_magnitudes == pytest.approx(np.array([1.0, 1.0]), 1.0e-4)

        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.GridIrregularGrouped([(2.0, 0.0), (0.0, 2.0)])

        deflection_magnitudes = sis.deflection_magnitudes_from_grid(grid=grid)

        assert deflection_magnitudes == pytest.approx(np.array([2.0, 2.0]), 1.0e-4)

        grid = ag.Grid.uniform(shape_2d=(5, 5), pixel_scales=0.1, sub_size=1)

        deflections = sis.deflections_from_grid(grid=grid)
        magitudes_manual = np.sqrt(
            np.square(deflections[:, 0]) + np.square(deflections[:, 1])
        )

        deflection_magnitudes = sis.deflection_magnitudes_from_grid(grid=grid)

        assert deflection_magnitudes == pytest.approx(magitudes_manual, 1.0e-4)


class TestDeflectionsViaPotential:
    def test__compare_sis_deflections_via_potential_and_calculation(self):
        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(10, 10), pixel_scales=0.05, sub_size=1)

        deflections_via_calculation = sis.deflections_from_grid(grid=grid)

        deflections_via_potential = sis.deflections_via_potential_from_grid(grid=grid)

        mean_error = np.mean(
            deflections_via_potential.in_1d - deflections_via_calculation.in_1d
        )

        assert mean_error < 1e-4

    def test__compare_sie_at_phi_45__deflections_via_potential_and_calculation(self):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.111111, 0.0), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(10, 10), pixel_scales=0.05, sub_size=1)

        deflections_via_calculation = sie.deflections_from_grid(grid=grid)

        deflections_via_potential = sie.deflections_via_potential_from_grid(grid=grid)

        mean_error = np.mean(
            deflections_via_potential.in_1d - deflections_via_calculation.in_1d
        )

        assert mean_error < 1e-4

    def test__compare_sie_at_phi_0__deflections_via_potential_and_calculation(self):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.0, -0.111111), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(10, 10), pixel_scales=0.05, sub_size=1)

        deflections_via_calculation = sie.deflections_from_grid(grid=grid)

        deflections_via_potential = sie.deflections_via_potential_from_grid(grid=grid)

        mean_error = np.mean(
            deflections_via_potential.in_1d - deflections_via_calculation.in_1d
        )

        assert mean_error < 1e-4


class TestJacobian:
    def test__jacobian_components(self):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.0, -0.111111), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(100, 100), pixel_scales=0.05, sub_size=1)

        jacobian = sie.jacobian_from_grid(grid=grid)

        A_12 = jacobian[0][1]
        A_21 = jacobian[1][0]

        mean_error = np.mean(A_12.in_1d - A_21.in_1d)

        assert mean_error < 1e-4

        grid = ag.Grid.uniform(shape_2d=(100, 100), pixel_scales=0.05, sub_size=2)

        jacobian = sie.jacobian_from_grid(grid=grid)

        A_12 = jacobian[0][1]
        A_21 = jacobian[1][0]

        mean_error = np.mean(A_12.in_1d - A_21.in_1d)

        assert mean_error < 1e-4


class TestMagnification:
    def test__compare_magnification_from_eigen_values_and_from_determinant(self):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.0, -0.111111), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(100, 100), pixel_scales=0.05, sub_size=1)

        magnification_via_determinant = sie.magnification_from_grid(grid=grid)

        tangential_eigen_value = sie.tangential_eigen_value_from_grid(grid=grid)

        radal_eigen_value = sie.radial_eigen_value_from_grid(grid=grid)

        magnification_via_eigen_values = 1 / (
            tangential_eigen_value * radal_eigen_value
        )

        mean_error = np.mean(
            magnification_via_determinant.in_1d - magnification_via_eigen_values.in_1d
        )

        assert mean_error < 1e-4

        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.0, -0.111111), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(100, 100), pixel_scales=0.05, sub_size=2)

        magnification_via_determinant = sie.magnification_from_grid(grid=grid)

        tangential_eigen_value = sie.tangential_eigen_value_from_grid(grid=grid)

        radal_eigen_value = sie.radial_eigen_value_from_grid(grid=grid)

        magnification_via_eigen_values = 1 / (
            tangential_eigen_value * radal_eigen_value
        )

        mean_error = np.mean(
            magnification_via_determinant.in_1d - magnification_via_eigen_values.in_1d
        )

        assert mean_error < 1e-4

    def test__compare_magnification_from_determinant_and_from_convergence_and_shear(
        self,
    ):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.0, -0.111111), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(100, 100), pixel_scales=0.05, sub_size=1)

        magnification_via_determinant = sie.magnification_from_grid(grid=grid)

        convergence = sie.convergence_via_jacobian_from_grid(grid=grid)

        shear = sie.shear_via_jacobian_from_grid(grid=grid)

        magnification_via_convergence_and_shear = 1 / (
            (1 - convergence) ** 2 - shear ** 2
        )

        mean_error = np.mean(
            magnification_via_determinant.in_1d
            - magnification_via_convergence_and_shear.in_1d
        )

        assert mean_error < 1e-4

        grid = ag.Grid.uniform(shape_2d=(100, 100), pixel_scales=0.05, sub_size=2)

        magnification_via_determinant = sie.magnification_from_grid(grid=grid)

        convergence = sie.convergence_via_jacobian_from_grid(grid=grid)

        shear = sie.shear_via_jacobian_from_grid(grid=grid)

        magnification_via_convergence_and_shear = 1 / (
            (1 - convergence) ** 2 - shear ** 2
        )

        mean_error = np.mean(
            magnification_via_determinant.in_1d
            - magnification_via_convergence_and_shear.in_1d
        )

        assert mean_error < 1e-4

    def test__magnification_irregular_from_grid(self):

        # grid = ag.Grid.uniform(shape_2d=(100, 100), pixel_scales=0.05, sub_size=1)
        #
        # print(grid.in_2d[28, 47])
        # print(grid.in_2d[67, 48])
        # print(grid.in_2d[68, 48])
        #
        # sis = MockEllipticalIsothermal(
        #     centre=(0.0, 0.0),
        #     elliptical_comps=(0.001, 0.001),
        #     einstein_radius=1.0
        # )
        #
        # print(sis.magnification_from_grid(grid=grid).in_2d[28, 47])
        # print(sis.magnification_from_grid(grid=grid).in_2d[67, 48])
        # print(sis.magnification_from_grid(grid=grid).in_2d[68, 48])
        #
        # print()
        #
        # buffer = 0.001
        #
        # print(sis.magnification_irregular_from_grid(grid=ag.GridIrregularGrouped(grid=[[(1.100039, -0.00742)]]), buffer=buffer))
        # print(sis.magnification_irregular_from_grid(grid=ag.GridIrregularGrouped(grid=[[(-0.90039, -0.00585)]]), buffer=buffer))
        # print(sis.magnification_irregular_from_grid(grid=ag.GridIrregularGrouped(grid=[[(-0.95039, -0.00585)]]), buffer=buffer))
        # print(sis.magnification_irregular_from_grid(grid=ag.GridIrregularGrouped(grid=[[(-1.0, -0.00585)]]), buffer=buffer))
        #
        # stop

        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.0, -0.111111), einstein_radius=2.0
        )

        grid = ag.GridIrregularGrouped(grid=[[(0.5, 0.5)], [(1.0, 1.0)]])

        magnification = sie.magnification_irregular_from_grid(grid=grid)

        assert magnification.in_grouped_list[0][0] == pytest.approx(-0.56303, 1.0e-4)
        assert magnification.in_grouped_list[1][0] == pytest.approx(-2.57591, 1.0e-4)


def critical_curve_via_magnification_from(mass_profile, grid):
    magnification = mass_profile.magnification_from_grid(grid=grid)

    inverse_magnification = 1 / magnification

    critical_curves_indices = measure.find_contours(inverse_magnification.in_2d, 0)

    no_critical_curves = len(critical_curves_indices)
    contours = []
    critical_curves = []

    for jj in np.arange(no_critical_curves):
        contours.append(critical_curves_indices[jj])
        contour_x, contour_y = contours[jj].T
        pixel_coord = np.stack((contour_x, contour_y), axis=-1)

        critical_curve = grid.geometry.grid_scaled_from_grid_pixels_1d_for_marching_squares(
            grid_pixels_1d=pixel_coord, shape_2d=magnification.sub_shape_2d
        )

        critical_curves.append(critical_curve)

    return critical_curves


def caustics_via_magnification_from(mass_profile, grid):
    caustics = []

    critical_curves = critical_curve_via_magnification_from(
        mass_profile=mass_profile, grid=grid
    )

    for i in range(len(critical_curves)):
        critical_curve = critical_curves[i]

        deflections_1d = mass_profile.deflections_from_grid(grid=critical_curve)

        caustic = critical_curve - deflections_1d

        caustics.append(caustic)

    return caustics


class TestConvergenceViajacobian:
    def test__compare_sis_convergence_via_jacobian_and_calculation(self):
        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(20, 20), pixel_scales=0.05, sub_size=1)

        convergence_via_calculation = sis.convergence_from_grid(grid=grid)

        convergence_via_jacobian = sis.convergence_via_jacobian_from_grid(grid=grid)

        mean_error = np.mean(
            convergence_via_jacobian.in_1d - convergence_via_calculation.in_1d
        )

        assert convergence_via_jacobian.in_2d_binned.shape == (20, 20)
        assert mean_error < 1e-1

        mean_error = np.mean(
            convergence_via_jacobian.in_1d - convergence_via_calculation.in_1d
        )

        assert mean_error < 1e-1

    def test__compare_sie_at_phi_45__convergence_via_jacobian_and_calculation(self):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.111111, 0.0), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(20, 20), pixel_scales=0.05, sub_size=1)

        convergence_via_calculation = sie.convergence_from_grid(grid=grid)

        convergence_via_jacobian = sie.convergence_via_jacobian_from_grid(grid=grid)

        mean_error = np.mean(
            convergence_via_jacobian.in_1d - convergence_via_calculation.in_1d
        )

        assert mean_error < 1e-1


class TestEvaluationGrid:
    def test__grid_changes_resolution_based_on_pixel_scale_input(self):
        @lensing.evaluation_grid
        def mock_func(lensing_obj, grid, pixel_scale=0.05):
            return grid

        grid = ag.Grid.uniform(shape_2d=(4, 4), pixel_scales=0.05)

        evaluation_grid = mock_func(lensing_obj=None, grid=grid, pixel_scale=0.05)

        assert (evaluation_grid == grid).all()

        evaluation_grid = mock_func(lensing_obj=None, grid=grid, pixel_scale=0.1)
        downscaled_grid = ag.Grid.uniform(shape_2d=(2, 2), pixel_scales=0.1)

        assert (evaluation_grid == downscaled_grid).all()

        evaluation_grid = mock_func(lensing_obj=None, grid=grid, pixel_scale=0.025)
        upscaled_grid = ag.Grid.uniform(shape_2d=(8, 8), pixel_scales=0.025)

        assert (evaluation_grid == upscaled_grid).all()

        evaluation_grid = mock_func(lensing_obj=None, grid=grid, pixel_scale=0.03)
        upscaled_grid = ag.Grid.uniform(shape_2d=(6, 6), pixel_scales=0.03)

        assert (evaluation_grid == upscaled_grid).all()

    def test__grid_changes_to_uniform_and_zoomed_in_if_masked(self):
        @lensing.evaluation_grid
        def mock_func(lensing_obj, grid, pixel_scale=0.05):
            return grid

        mask = ag.Mask2D.circular(shape_2d=(11, 11), pixel_scales=1.0, radius=3.0)

        grid = ag.Grid.from_mask(mask=mask)

        evaluation_grid = mock_func(lensing_obj=None, grid=grid, pixel_scale=1.0)
        grid_uniform = ag.Grid.uniform(shape_2d=(7, 7), pixel_scales=1.0)

        assert (evaluation_grid[0] == np.array([3.0, -3.0])).all()
        assert (evaluation_grid == grid_uniform).all()

        mask = ag.Mask2D.circular(
            shape_2d=(29, 29), pixel_scales=1.0, radius=3.0, centre=(5.0, 5.0)
        )

        grid = ag.Grid.from_mask(mask=mask)

        evaluation_grid = mock_func(lensing_obj=None, grid=grid, pixel_scale=1.0)
        grid_uniform = ag.Grid.uniform(
            shape_2d=(7, 7), pixel_scales=1.0, origin=(5.0, 5.0)
        )

        assert (evaluation_grid[0] == np.array([8.0, 2.0])).all()
        assert (evaluation_grid == grid_uniform).all()


class TestCriticalCurvesAndCaustics:
    def test_compare_magnification_from_determinant_and_from_convergence_and_shear(
        self,
    ):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.0, -0.111111), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(100, 100), pixel_scales=0.05, sub_size=2)

        magnification_via_determinant = sie.magnification_from_grid(grid=grid)

        convergence = sie.convergence_via_jacobian_from_grid(grid=grid)

        shear = sie.shear_via_jacobian_from_grid(grid=grid)

        magnification_via_convergence_and_shear = 1 / (
            (1 - convergence) ** 2 - shear ** 2
        )

        mean_error = np.mean(
            magnification_via_determinant - magnification_via_convergence_and_shear
        )

        assert mean_error < 1e-2

    def test__tangential_critical_curve_radii__spherical_isothermal(self):

        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(15, 15), pixel_scales=0.3)

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        tangential_critical_curve = np.asarray(critical_curves.in_grouped_list[0])

        x_critical_tangential, y_critical_tangential = (
            tangential_critical_curve[:, 1],
            tangential_critical_curve[:, 0],
        )

        assert np.mean(
            x_critical_tangential ** 2 + y_critical_tangential ** 2
        ) == pytest.approx(sis.einstein_radius ** 2, 5e-1)

    def test__tangential_critical_curve_centres__spherical_isothermal(self):
        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        tangential_critical_curve = np.asarray(critical_curves.in_grouped_list[0])

        y_centre = np.mean(tangential_critical_curve[:, 0])
        x_centre = np.mean(tangential_critical_curve[:, 1])

        assert -0.03 < y_centre < 0.03
        assert -0.03 < x_centre < 0.03

        sis = MockSphericalIsothermal(centre=(0.5, 1.0), einstein_radius=2.0)

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        tangential_critical_curve = np.asarray(critical_curves.in_grouped_list[0])

        y_centre = np.mean(tangential_critical_curve[:, 0])
        x_centre = np.mean(tangential_critical_curve[:, 1])

        assert 0.47 < y_centre < 0.53
        assert 0.97 < x_centre < 1.03

    def test__radial_critical_curve_centres__spherical_isothermal(self):

        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        radial_critical_curve = np.asarray(critical_curves.in_grouped_list[1])

        y_centre = np.mean(radial_critical_curve[:, 0])
        x_centre = np.mean(radial_critical_curve[:, 1])

        assert -0.05 < y_centre < 0.05
        assert -0.05 < x_centre < 0.05

        sis = MockSphericalIsothermal(centre=(0.5, 1.0), einstein_radius=2.0)

        critical_curves = sis.critical_curves_from_grid(grid=grid)

        radial_critical_curve = np.asarray(critical_curves.in_grouped_list[1])

        y_centre = np.mean(radial_critical_curve[:, 0])
        x_centre = np.mean(radial_critical_curve[:, 1])

        assert 0.45 < y_centre < 0.55
        assert 0.95 < x_centre < 1.05

    def test__tangential_caustic_centres__spherical_isothermal(self):
        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        caustics = sis.caustics_from_grid(grid=grid)

        tangential_caustic = np.asarray(caustics.in_grouped_list[0])

        y_centre = np.mean(tangential_caustic[:, 0])
        x_centre = np.mean(tangential_caustic[:, 1])

        assert -0.03 < y_centre < 0.03
        assert -0.03 < x_centre < 0.03

        sis = MockSphericalIsothermal(centre=(0.5, 1.0), einstein_radius=2.0)

        caustics = sis.caustics_from_grid(grid=grid)

        tangential_caustic = np.asarray(caustics.in_grouped_list[0])

        y_centre = np.mean(tangential_caustic[:, 0])
        x_centre = np.mean(tangential_caustic[:, 1])

        assert 0.47 < y_centre < 0.53
        assert 0.97 < x_centre < 1.03

    def test__radial_caustics_radii__spherical_isothermal(self):
        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(20, 20), pixel_scales=0.2)

        caustics = sis.caustics_from_grid(grid=grid)

        caustic_radial = np.asarray(caustics.in_grouped_list[1])

        x_caustic_radial, y_caustic_radial = (
            caustic_radial[:, 1],
            caustic_radial[:, 0],
        )

        assert np.mean(x_caustic_radial ** 2 + y_caustic_radial ** 2) == pytest.approx(
            sis.einstein_radius ** 2, 5e-1
        )

    def test__radial_caustic_centres__spherical_isothermal(self):
        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        caustics = sis.caustics_from_grid(grid=grid)

        radial_caustic = np.asarray(caustics.in_grouped_list[1])

        y_centre = np.mean(radial_caustic[:, 0])
        x_centre = np.mean(radial_caustic[:, 1])

        assert -0.2 < y_centre < 0.2
        assert -0.35 < x_centre < 0.35

        sis = MockSphericalIsothermal(centre=(0.5, 1.0), einstein_radius=2.0)

        caustics = sis.caustics_from_grid(grid=grid)

        radial_caustic = np.asarray(caustics.in_grouped_list[1])

        y_centre = np.mean(radial_caustic[:, 0])
        x_centre = np.mean(radial_caustic[:, 1])

        assert 0.3 < y_centre < 0.7
        assert 0.7 < x_centre < 1.2

    def test__compare_tangential_critical_curves_from_magnification_and_eigen_values(
        self,
    ):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2, elliptical_comps=(0.109423, -0.019294)
        )

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        tangential_critical_curve_from_magnification = critical_curve_via_magnification_from(
            mass_profile=sie, grid=grid
        )[
            0
        ]

        tangential_critical_curve = sie.tangential_critical_curve_from_grid(
            grid=grid, pixel_scale=0.2
        )

        assert tangential_critical_curve == pytest.approx(
            tangential_critical_curve_from_magnification, 5e-1
        )

        tangential_critical_curve_from_magnification = critical_curve_via_magnification_from(
            mass_profile=sie, grid=grid
        )[
            0
        ]

        tangential_critical_curve = sie.tangential_critical_curve_from_grid(
            grid=grid, pixel_scale=0.2
        )

        assert tangential_critical_curve == pytest.approx(
            tangential_critical_curve_from_magnification, 5e-1
        )

    def test__compare_radial_critical_curves_from_magnification_and_eigen_values(self):

        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2, elliptical_comps=(0.109423, -0.019294)
        )

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        critical_curve_radial_from_magnification = critical_curve_via_magnification_from(
            mass_profile=sie, grid=grid
        )[
            1
        ]

        radial_critical_curve = sie.radial_critical_curve_from_grid(grid=grid)

        assert sum(critical_curve_radial_from_magnification) == pytest.approx(
            sum(radial_critical_curve), abs=0.7
        )

    def test__compare_tangential_caustic_from_magnification_and_eigen_values(self):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2, elliptical_comps=(0.109423, -0.019294)
        )

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        tangential_caustic_from_magnification = caustics_via_magnification_from(
            mass_profile=sie, grid=grid
        )[0]

        tangential_caustic = sie.tangential_caustic_from_grid(
            grid=grid, pixel_scale=0.2
        )

        assert sum(tangential_caustic) == pytest.approx(
            sum(tangential_caustic_from_magnification), 5e-1
        )

    def test__compare_radial_caustic_from_magnification_and_eigen_values__grid(self):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2, elliptical_comps=(0.109423, -0.019294)
        )

        grid = ag.Grid.uniform(shape_2d=(60, 60), pixel_scales=0.08)

        caustic_radial_from_magnification = caustics_via_magnification_from(
            mass_profile=sie, grid=grid
        )[1]

        radial_caustic = sie.radial_caustic_from_grid(grid=grid, pixel_scale=0.08)

        assert sum(radial_caustic) == pytest.approx(
            sum(caustic_radial_from_magnification), 7e-1
        )


class TestEinsteinRadiusMass:
    def test__tangential_critical_curve_area_from_critical_curve_and_calculation__spherical_isothermal(
        self,
    ):
        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        area_calc = np.pi * sis.einstein_radius ** 2

        area_within_tangential_critical_curve = sis.area_within_tangential_critical_curve_from_grid(
            grid=grid
        )

        assert area_within_tangential_critical_curve == pytest.approx(area_calc, 1e-1)

    def test__einstein_radius_from_tangential_critical_curve_values(self):

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        einstein_radius = sis.einstein_radius_from_grid(grid=grid)

        assert einstein_radius == pytest.approx(2.0, 1e-1)

        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), einstein_radius=2.0, elliptical_comps=(0.0, -0.25)
        )

        einstein_radius = sie.einstein_radius_from_grid(grid=grid)

        assert einstein_radius == pytest.approx(1.9360, 1e-1)

    def test__einstein_mass_from_tangential_critical_curve_values(self):

        grid = ag.Grid.uniform(shape_2d=(50, 50), pixel_scales=0.2)

        sis = MockSphericalIsothermal(centre=(0.0, 0.0), einstein_radius=2.0)

        einstein_mass = sis.einstein_mass_angular_from_grid(grid=grid)

        assert einstein_mass == pytest.approx(np.pi * 2.0 ** 2.0, 1e-1)


class TestGridBinning:
    def test__binning_works_on_all_from_grid_methods(self):
        sie = MockEllipticalIsothermal(
            centre=(0.0, 0.0), elliptical_comps=(0.0, -0.111111), einstein_radius=2.0
        )

        grid = ag.Grid.uniform(shape_2d=(10, 10), pixel_scales=0.05, sub_size=2)

        deflections = sie.deflections_via_potential_from_grid(grid=grid)

        deflections_first_binned_pixel = (
            deflections[0] + deflections[1] + deflections[2] + deflections[3]
        ) / 4

        assert deflections.in_1d_binned[0] == pytest.approx(
            deflections_first_binned_pixel, 1e-4
        )

        deflections_100th_binned_pixel = (
            deflections[399] + deflections[398] + deflections[397] + deflections[396]
        ) / 4

        assert deflections.in_1d_binned[99] == pytest.approx(
            deflections_100th_binned_pixel, 1e-4
        )

        jacobian = sie.jacobian_from_grid(grid=grid)

        jacobian_1st_pixel_binned_up = (
            jacobian[0][0][0]
            + jacobian[0][0][1]
            + jacobian[0][0][2]
            + jacobian[0][0][3]
        ) / 4

        assert jacobian[0][0].in_2d_binned.shape == (10, 10)
        assert jacobian[0][0].sub_shape_2d == (20, 20)
        assert jacobian[0][0].in_1d_binned[0] == pytest.approx(
            jacobian_1st_pixel_binned_up, 1e-4
        )

        jacobian_last_pixel_binned_up = (
            jacobian[0][0][399]
            + jacobian[0][0][398]
            + jacobian[0][0][397]
            + jacobian[0][0][396]
        ) / 4

        assert jacobian[0][0].in_1d_binned[99] == pytest.approx(
            jacobian_last_pixel_binned_up, 1e-4
        )

        shear_via_jacobian = sie.shear_via_jacobian_from_grid(grid=grid)

        shear_1st_pixel_binned_up = (
            shear_via_jacobian[0]
            + shear_via_jacobian[1]
            + shear_via_jacobian[2]
            + shear_via_jacobian[3]
        ) / 4

        assert shear_via_jacobian.in_1d_binned[0] == pytest.approx(
            shear_1st_pixel_binned_up, 1e-4
        )

        shear_last_pixel_binned_up = (
            shear_via_jacobian[399]
            + shear_via_jacobian[398]
            + shear_via_jacobian[397]
            + shear_via_jacobian[396]
        ) / 4

        assert shear_via_jacobian.in_1d_binned[99] == pytest.approx(
            shear_last_pixel_binned_up, 1e-4
        )

        tangential_eigen_values = sie.tangential_eigen_value_from_grid(grid=grid)

        first_pixel_binned_up = (
            tangential_eigen_values[0]
            + tangential_eigen_values[1]
            + tangential_eigen_values[2]
            + tangential_eigen_values[3]
        ) / 4

        assert tangential_eigen_values.in_1d_binned[0] == pytest.approx(
            first_pixel_binned_up, 1e-4
        )

        pixel_10000_from_av_sub_grid = (
            tangential_eigen_values[399]
            + tangential_eigen_values[398]
            + tangential_eigen_values[397]
            + tangential_eigen_values[396]
        ) / 4

        assert tangential_eigen_values.in_1d_binned[99] == pytest.approx(
            pixel_10000_from_av_sub_grid, 1e-4
        )

        radial_eigen_values = sie.radial_eigen_value_from_grid(grid=grid)

        first_pixel_binned_up = (
            radial_eigen_values[0]
            + radial_eigen_values[1]
            + radial_eigen_values[2]
            + radial_eigen_values[3]
        ) / 4

        assert radial_eigen_values.in_1d_binned[0] == pytest.approx(
            first_pixel_binned_up, 1e-4
        )

        pixel_10000_from_av_sub_grid = (
            radial_eigen_values[399]
            + radial_eigen_values[398]
            + radial_eigen_values[397]
            + radial_eigen_values[396]
        ) / 4

        assert radial_eigen_values.in_1d_binned[99] == pytest.approx(
            pixel_10000_from_av_sub_grid, 1e-4
        )



from autoarray.plot.plotters import abstract_plotters
from autoarray.plot.mat_wrap import mat_plot as mp
from autogalaxy.plot.mat_wrap import lensing_mat_plot, lensing_include, lensing_visuals


class HyperPlotter(abstract_plotters.AbstractPlotter):
    def __init__(
        self,
        mat_plot_2d: lensing_mat_plot.MatPlot2D = lensing_mat_plot.MatPlot2D(),
        visuals_2d: lensing_visuals.Visuals2D = lensing_visuals.Visuals2D(),
        include_2d: lensing_include.Include2D = lensing_include.Include2D(),
    ):
        super().__init__(
            mat_plot_2d=mat_plot_2d, include_2d=include_2d, visuals_2d=visuals_2d
        )

    @property
    def visuals_with_include_2d(self) -> lensing_visuals.Visuals2D:
        """
        Extracts from a `Structure` attributes that can be plotted and return them in a `Visuals` object.

        Only attributes with `True` entries in the `Include` object are extracted for plotting.

        From an `AbstractStructure` the following attributes can be extracted for plotting:

        - origin: the (y,x) origin of the structure's coordinate system.
        - mask: the mask of the structure.
        - border: the border of the structure's mask.

        Parameters
        ----------
        structure : abstract_structure.AbstractStructure
            The structure whose attributes are extracted for plotting.

        Returns
        -------
        vis.Visuals2D
            The collection of attributes that can be plotted by a `Plotter2D` object.
        """
        return self.visuals_2d

    def figure_hyper_model_image(self, hyper_model_image):
        """Plot the image of a hyper_galaxies galaxy image.

        Set *autogalaxy.datas.arrays.mat_plot_2d.mat_plot_2d* for a description of all input parameters not described below.

        Parameters
        -----------
        hyper_galaxy_image : datas.imaging.datas.Imaging
            The hyper_galaxies galaxy image.
        origin : True
            If true, the origin of the datas's coordinate system is plotted as a 'x'.
        """
        self.mat_plot_2d.plot_array(
            array=hyper_model_image,
            visuals_2d=self.visuals_with_include_2d,
            auto_labels=mp.AutoLabels(
                title="Hyper Model Image", filename="hyper_model_image"
            ),
        )

    def figure_hyper_galaxy_image(self, galaxy_image):
        """Plot the image of a hyper_galaxies galaxy image.

        Set *autogalaxy.datas.arrays.mat_plot_2d.mat_plot_2d* for a description of all input parameters not described below.

        Parameters
        -----------
        hyper_galaxy_image : datas.imaging.datas.Imaging
            The hyper_galaxies galaxy image.
        origin : True
            If true, the origin of the datas's coordinate system is plotted as a 'x'.
        """
        self.mat_plot_2d.plot_array(
            array=galaxy_image,
            visuals_2d=self.visuals_with_include_2d,
            auto_labels=mp.AutoLabels(
                title="Hyper Galaxy Image", filename="hyper_galaxy_image"
            ),
        )

    def figure_contribution_map(self, contribution_map_in):
        """Plot the summed contribution maps of a hyper_galaxies-fit.

        Set *autogalaxy.datas.arrays.mat_plot_2d.mat_plot_2d* for a description of all input parameters not described below.

        Parameters
        -----------
        fit : datas.fitting.fitting.AbstractLensHyperFit
            The hyper_galaxies-fit to the datas, which includes a list of every model image, residual_map, chi-squareds, etc.
        image_index : int
            The index of the datas in the datas-set of which the contribution_maps are plotted.
        """
        self.mat_plot_2d.plot_array(
            array=contribution_map_in,
            visuals_2d=self.visuals_with_include_2d,
            auto_labels=mp.AutoLabels(
                title="Contribution Map", filename="contribution_map"
            ),
        )

    def subplot_hyper_images_of_galaxies(self, hyper_galaxy_image_path_dict):

        if hyper_galaxy_image_path_dict is None:
            return

        self.open_subplot_figure(number_subplots=len(hyper_galaxy_image_path_dict))

        for path, galaxy_image in hyper_galaxy_image_path_dict.items():

            self.figure_hyper_galaxy_image(galaxy_image=galaxy_image)

        self.mat_plot_2d.output.subplot_to_figure(
            auto_filename="subplot_hyper_images_of_galaxies"
        )

        self.close_subplot_figure()

    def subplot_contribution_maps_of_galaxies(self, contribution_maps_of_galaxies):

        contribution_maps = [
            contribution_map
            for contribution_map in contribution_maps_of_galaxies
            if contribution_map is not None
        ]

        number_subplots = len(contribution_maps)

        if number_subplots == 0:
            return

        self.open_subplot_figure(number_subplots=number_subplots)

        for contribution_map_array in contribution_maps:

            self.figure_contribution_map(contribution_map_in=contribution_map_array)

        self.mat_plot_2d.output.subplot_to_figure(
            auto_filename="subplot_contribution_maps_of_galaxies"
        )

        self.close_subplot_figure()



import math

import numpy as np
from astropy import constants


def arcsec_per_kpc_from(*, redshift, cosmology):
    return cosmology.arcsec_per_kpc_proper(z=redshift).value


def kpc_per_arcsec_from(*, redshift, cosmology):
    return 1.0 / cosmology.arcsec_per_kpc_proper(z=redshift).value


def angular_diameter_distance_to_earth_in_kpc_from(*, redshift, cosmology):

    angular_diameter_distance_kpc = cosmology.angular_diameter_distance(z=redshift).to(
        "kpc"
    )

    return angular_diameter_distance_kpc.value


def angular_diameter_distance_between_redshifts_in_kpc_from(
    *, redshift_0, redshift_1, cosmology
):

    angular_diameter_distance_between_redshifts_kpc = cosmology.angular_diameter_distance_z1z2(
        redshift_0, redshift_1
    ).to(
        "kpc"
    )

    return angular_diameter_distance_between_redshifts_kpc.value


def cosmic_average_density_from(*, redshift, cosmology):

    cosmic_average_density_kpc = (
        cosmology.critical_density(z=redshift).to("solMass / kpc^3").value
    )

    kpc_per_arcsec = kpc_per_arcsec_from(redshift=redshift, cosmology=cosmology)

    return cosmic_average_density_kpc * kpc_per_arcsec ** 3.0


def cosmic_average_density_solar_mass_per_kpc3_from(*, redshift, cosmology):

    cosmic_average_density_kpc = (
        cosmology.critical_density(z=redshift).to("solMass / kpc^3").value
    )

    return cosmic_average_density_kpc


def critical_surface_density_between_redshifts_from(
    *, redshift_0, redshift_1, cosmology
):

    critical_surface_density_kpc = critical_surface_density_between_redshifts_solar_mass_per_kpc2_from(
        redshift_0=redshift_0, redshift_1=redshift_1, cosmology=cosmology
    )

    kpc_per_arcsec = kpc_per_arcsec_from(redshift=redshift_0, cosmology=cosmology)

    return critical_surface_density_kpc * kpc_per_arcsec ** 2.0


def critical_surface_density_between_redshifts_solar_mass_per_kpc2_from(
    *, redshift_0, redshift_1, cosmology
):

    const = constants.c.to("kpc / s") ** 2.0 / (
        4 * math.pi * constants.G.to("kpc3 / (solMass s2)")
    )

    angular_diameter_distance_of_redshift_0_to_earth_kpc = angular_diameter_distance_to_earth_in_kpc_from(
        redshift=redshift_0, cosmology=cosmology
    )

    angular_diameter_distance_of_redshift_1_to_earth_kpc = angular_diameter_distance_to_earth_in_kpc_from(
        redshift=redshift_1, cosmology=cosmology
    )

    angular_diameter_distance_between_redshifts_kpc = angular_diameter_distance_between_redshifts_in_kpc_from(
        redshift_0=redshift_0, redshift_1=redshift_1, cosmology=cosmology
    )

    return (
        const
        * angular_diameter_distance_of_redshift_1_to_earth_kpc
        / (
            angular_diameter_distance_between_redshifts_kpc
            * angular_diameter_distance_of_redshift_0_to_earth_kpc
        )
    ).value


def scaling_factor_between_redshifts_from(
    *, redshift_0, redshift_1, redshift_final, cosmology
):

    angular_diameter_distance_between_redshifts_0_and_1 = (
        cosmology.angular_diameter_distance_z1z2(z1=redshift_0, z2=redshift_1)
        .to("kpc")
        .value
    )

    angular_diameter_distance_to_redshift_final = (
        cosmology.angular_diameter_distance(z=redshift_final).to("kpc").value
    )

    angular_diameter_distance_of_redshift_1_to_earth = (
        cosmology.angular_diameter_distance(z=redshift_1).to("kpc").value
    )

    angular_diameter_distance_between_redshift_1_and_final = (
        cosmology.angular_diameter_distance_z1z2(z1=redshift_0, z2=redshift_final)
        .to("kpc")
        .value
    )

    return (
        angular_diameter_distance_between_redshifts_0_and_1
        * angular_diameter_distance_to_redshift_final
    ) / (
        angular_diameter_distance_of_redshift_1_to_earth
        * angular_diameter_distance_between_redshift_1_and_final
    )


def velocity_dispersion_from(*, redshift_0, redshift_1, einstein_radius, cosmology):

    const = constants.c.to("kpc / s")

    angular_diameter_distance_to_redshift_0_kpc = angular_diameter_distance_to_earth_in_kpc_from(
        redshift=redshift_1, cosmology=cosmology
    )

    angular_diameter_distance_to_redshift_1_kpc = angular_diameter_distance_to_earth_in_kpc_from(
        redshift=redshift_1, cosmology=cosmology
    )

    angular_diameter_distance_between_redshifts_kpc = angular_diameter_distance_between_redshifts_in_kpc_from(
        redshift_0=redshift_0, redshift_1=redshift_1, cosmology=cosmology
    )

    kpc_per_arcsec = kpc_per_arcsec_from(redshift=redshift_0, cosmology=cosmology)

    einstein_radius_kpc = einstein_radius * kpc_per_arcsec

    velocity_dispersion_kpc = const * np.sqrt(
        (einstein_radius_kpc * angular_diameter_distance_to_redshift_1_kpc)
        / (
            4
            * np.pi
            * angular_diameter_distance_to_redshift_0_kpc
            * angular_diameter_distance_between_redshifts_kpc
        )
    )

    return velocity_dispersion_kpc.to("km/s").value



from os import path

import pytest

import autogalaxy as ag
from autogalaxy.mock import mock

pytestmark = pytest.mark.filterwarnings(
    "ignore:Using a non-tuple sequence for multidimensional indexing is deprecated; use `arr[tuple(seq)]` instead of "
    "`arr[seq]`. In the future this will be interpreted as an arrays index, `arr[np.arrays(seq)]`, which will result "
    "either in an error or a different result."
)

directory = path.dirname(path.realpath(__file__))


class TestPhase:
    def test__extend_with_hyper_phase(self):

        phase_with_hyper_sky = ag.PhaseImaging(search=mock.MockSearch())

        setup_hyper = ag.SetupHyper(
            hyper_image_sky=ag.hyper_data.HyperImageSky,
            hyper_search_with_inversion=mock.MockSearch("test_phase"),
        )

        phase_extended = phase_with_hyper_sky.extend_with_hyper_phase(
            setup_hyper=setup_hyper
        )

        assert isinstance(phase_extended, ag.HyperPhase)
        assert phase_extended.hyper_image_sky is ag.hyper_data.HyperImageSky

        phase_with_hyper_sky = ag.PhaseImaging(search=mock.MockSearch())

        phase_extended = phase_with_hyper_sky.extend_with_hyper_phase(
            setup_hyper=ag.SetupHyper(
                hyper_background_noise=ag.hyper_data.HyperBackgroundNoise,
                hyper_search_with_inversion=mock.MockSearch("test_phase"),
            )
        )

        assert isinstance(phase_extended, ag.HyperPhase)

        phase_with_pixelization = ag.PhaseImaging(
            galaxies=dict(
                source=ag.GalaxyModel(
                    redshift=0.5,
                    pixelization=ag.pix.Rectangular,
                    regularization=ag.reg.Constant,
                )
            ),
            search=mock.MockSearch(),
        )

        phase_extended = phase_with_pixelization.extend_with_hyper_phase(
            setup_hyper=ag.SetupHyper(
                hyper_search_with_inversion=mock.MockSearch("test_phase")
            )
        )

        assert isinstance(phase_extended, ag.HyperPhase)

    def test__extend_with_hyper_phase__does_not_extend_if_no_hyper_compoennts_to_optimize(
        self
    ):

        phase_no_pixelization = ag.PhaseImaging(search=mock.MockSearch("test_phase"))

        phase_extended = phase_no_pixelization.extend_with_hyper_phase(
            setup_hyper=ag.SetupHyper()
        )

        assert phase_extended == phase_no_pixelization

        phase_no_pixelization = ag.PhaseImaging(search=mock.MockSearch("test_phase"))

        phase_extended = phase_no_pixelization.extend_with_hyper_phase(
            setup_hyper=ag.SetupHyper(hyper_image_sky=ag.hyper_data.HyperImageSky),
            include_hyper_image_sky=False,
        )

        assert phase_extended == phase_no_pixelization


class TestMakeAnalysis:
    def test__mask_input_uses_mask(self, phase_imaging_7x7, imaging_7x7):
        # If an input mask is supplied we use mask input.

        mask_input = ag.Mask2D.circular(
            shape_2d=imaging_7x7.shape_2d, pixel_scales=1.0, sub_size=1, radius=1.5
        )

        analysis = phase_imaging_7x7.make_analysis(
            dataset=imaging_7x7, mask=mask_input, results=mock.MockResults()
        )

        assert (analysis.masked_imaging.mask == mask_input).all()
        assert analysis.masked_imaging.mask.pixel_scales == mask_input.pixel_scales

    def test__mask_changes_sub_size_depending_on_phase_attribute(
        self, phase_imaging_7x7, imaging_7x7
    ):
        # If an input mask is supplied we use mask input.

        mask_input = ag.Mask2D.circular(
            shape_2d=imaging_7x7.shape_2d, pixel_scales=1, sub_size=1, radius=1.5
        )

        phase_imaging_7x7 = ag.PhaseImaging(
            search=mock.MockSearch("test_phase"),
            settings=ag.SettingsPhaseImaging(
                settings_masked_imaging=ag.SettingsMaskedImaging(sub_size=1)
            ),
        )

        analysis = phase_imaging_7x7.make_analysis(
            dataset=imaging_7x7, mask=mask_input, results=mock.MockResults()
        )

        assert (analysis.masked_imaging.mask == mask_input).all()
        assert analysis.masked_imaging.mask.sub_size == 1
        assert analysis.masked_imaging.mask.pixel_scales == mask_input.pixel_scales

        phase_imaging_7x7 = ag.PhaseImaging(
            search=mock.MockSearch("test_phase"),
            settings=ag.SettingsPhaseImaging(
                settings_masked_imaging=ag.SettingsMaskedImaging(sub_size=2)
            ),
        )

        analysis = phase_imaging_7x7.make_analysis(
            dataset=imaging_7x7, mask=mask_input, results=mock.MockResults()
        )

        assert (analysis.masked_imaging.mask == mask_input).all()
        assert analysis.masked_imaging.mask.sub_size == 2
        assert analysis.masked_imaging.mask.pixel_scales == mask_input.pixel_scales


class TestPhasePickle:

    # noinspection PyTypeChecker
    def test_assertion_failure(self, imaging_7x7, mask_7x7):
        phase_imaging_7x7 = ag.PhaseImaging(
            galaxies=dict(
                galaxy=ag.Galaxy(light=ag.lp.EllipticalLightProfile, redshift=1)
            ),
            search=mock.MockSearch("name"),
        )

        result = phase_imaging_7x7.run(dataset=imaging_7x7, mask=mask_7x7, results=None)
        assert result is not None

        phase_imaging_7x7 = ag.PhaseImaging(
            galaxies=dict(
                galaxy=ag.Galaxy(light=ag.lp.EllipticalLightProfile, redshift=1)
            ),
            search=mock.MockSearch("name"),
        )

        result = phase_imaging_7x7.run(dataset=imaging_7x7, mask=mask_7x7, results=None)
        assert result is not None



from os import path

import pytest

import autogalaxy.plot as aplt

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_fit_imaging_plotter_setup():
    return path.join(
        "{}".format(path.dirname(path.realpath(__file__))), "files", "plots", "fit"
    )


def test__fit_individuals__source_and_galaxy__dependent_on_input(
    masked_imaging_fit_x2_galaxy_7x7, include_2d_all, plot_path, plot_patch
):

    fit_imaging_plotter = aplt.FitImagingPlotter(
        fit=masked_imaging_fit_x2_galaxy_7x7,
        include_2d=include_2d_all,
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(plot_path, format="png")),
    )

    fit_imaging_plotter.figures(
        image=True,
        noise_map=False,
        signal_to_noise_map=False,
        model_image=True,
        chi_squared_map=True,
    )

    assert path.join(plot_path, "image.png") in plot_patch.paths
    assert path.join(plot_path, "noise_map.png") not in plot_patch.paths
    assert path.join(plot_path, "signal_to_noise_map.png") not in plot_patch.paths
    assert path.join(plot_path, "model_image.png") in plot_patch.paths
    assert path.join(plot_path, "residual_map.png") not in plot_patch.paths
    assert path.join(plot_path, "normalized_residual_map.png") not in plot_patch.paths
    assert path.join(plot_path, "chi_squared_map.png") in plot_patch.paths


def test__figures_of_galaxies(
    masked_imaging_fit_x2_galaxy_7x7, include_2d_all, plot_path, plot_patch
):

    fit_imaging_plotter = aplt.FitImagingPlotter(
        fit=masked_imaging_fit_x2_galaxy_7x7,
        include_2d=include_2d_all,
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(plot_path, format="png")),
    )

    fit_imaging_plotter.figures_of_galaxies(subtracted_image=True)

    assert path.join(plot_path, "subtracted_image_of_galaxy_0.png") in plot_patch.paths
    assert path.join(plot_path, "subtracted_image_of_galaxy_1.png") in plot_patch.paths

    fit_imaging_plotter.figures_of_galaxies(model_image=True)

    assert path.join(plot_path, "model_image_of_galaxy_0.png") in plot_patch.paths
    assert path.join(plot_path, "model_image_of_galaxy_1.png") in plot_patch.paths

    plot_patch.paths = []

    fit_imaging_plotter.figures_of_galaxies(subtracted_image=True, galaxy_index=0)

    assert path.join(plot_path, "subtracted_image_of_galaxy_0.png") in plot_patch.paths
    assert (
        path.join(plot_path, "subtracted_image_of_galaxy_1.png") not in plot_patch.paths
    )

    fit_imaging_plotter.figures_of_galaxies(model_image=True, galaxy_index=1)

    assert path.join(plot_path, "model_image_of_galaxy_0.png") not in plot_patch.paths
    assert path.join(plot_path, "model_image_of_galaxy_1.png") in plot_patch.paths


def test__subplot_of_galaxy(
    masked_imaging_fit_x2_galaxy_7x7, include_2d_all, plot_path, plot_patch
):

    fit_imaging_plotter = aplt.FitImagingPlotter(
        fit=masked_imaging_fit_x2_galaxy_7x7,
        include_2d=include_2d_all,
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(plot_path, format="png")),
    )
    fit_imaging_plotter.subplots_of_galaxies()
    assert path.join(plot_path, "subplot_of_galaxy_0.png") in plot_patch.paths
    assert path.join(plot_path, "subplot_of_galaxy_1.png") in plot_patch.paths

    plot_patch.paths = []

    fit_imaging_plotter.subplots_of_galaxies(galaxy_index=0)

    assert path.join(plot_path, "subplot_of_galaxy_0.png") in plot_patch.paths
    assert path.join(plot_path, "subplot_of_galaxy_1.png") not in plot_patch.paths



from os import path

import autogalaxy as ag
import autogalaxy.plot as aplt
import pytest

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_galaxy_plotter_setup():
    return path.join(
        "{}".format(path.dirname(path.realpath(__file__))), "files", "plots", "galaxy"
    )


def test__visuals_with_include_2d(gal_x1_lp_x1_mp, grid_7x7):

    visuals_2d = aplt.Visuals2D(vector_field=2)

    include = aplt.Include2D(
        origin=True,
        mask=True,
        border=True,
        light_profile_centres=True,
        mass_profile_centres=True,
        critical_curves=True,
        caustics=True,
    )

    galaxy_plotter = aplt.GalaxyPlotter(
        galaxy=gal_x1_lp_x1_mp, grid=grid_7x7, visuals_2d=visuals_2d, include_2d=include
    )

    assert galaxy_plotter.visuals_with_include_2d.origin.in_1d_list == [(0.0, 0.0)]
    assert (galaxy_plotter.visuals_with_include_2d.mask == grid_7x7.mask).all()
    assert (
        galaxy_plotter.visuals_with_include_2d.border
        == grid_7x7.mask.geometry.border_grid_sub_1.in_1d_binned
    ).all()
    assert galaxy_plotter.visuals_with_include_2d.light_profile_centres.in_1d_list == [
        gal_x1_lp_x1_mp.light_profile_0.centre
    ]
    assert galaxy_plotter.visuals_with_include_2d.mass_profile_centres.in_1d_list == [
        gal_x1_lp_x1_mp.mass_profile_0.centre
    ]
    assert galaxy_plotter.visuals_with_include_2d.vector_field == 2

    include = aplt.Include2D(origin=False, mask=False, border=False)

    galaxy_plotter = aplt.GalaxyPlotter(
        galaxy=gal_x1_lp_x1_mp, grid=grid_7x7, visuals_2d=visuals_2d, include_2d=include
    )

    assert galaxy_plotter.visuals_with_include_2d.origin == None
    assert galaxy_plotter.visuals_with_include_2d.mask == None
    assert galaxy_plotter.visuals_with_include_2d.border == None
    assert galaxy_plotter.visuals_with_include_2d.vector_field == 2


def test__individual_images_are_output(
    gal_x1_lp_x1_mp,
    sub_grid_7x7,
    mask_7x7,
    grid_irregular_grouped_7x7,
    include_2d_all,
    plot_path,
    plot_patch,
):

    galaxy_plotter = aplt.GalaxyPlotter(
        galaxy=gal_x1_lp_x1_mp,
        grid=sub_grid_7x7,
        include_2d=include_2d_all,
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(plot_path, format="png")),
    )
    galaxy_plotter.figures(image=True, convergence=True)

    assert path.join(plot_path, "image.png") in plot_patch.paths
    assert path.join(plot_path, "convergence.png") in plot_patch.paths

    gal_x1_lp_x1_mp.hyper_galaxy = ag.HyperGalaxy()
    gal_x1_lp_x1_mp.hyper_model_image = ag.Array.ones(shape_2d=(7, 7), pixel_scales=0.1)
    gal_x1_lp_x1_mp.hyper_galaxy_image = ag.Array.ones(
        shape_2d=(7, 7), pixel_scales=0.1
    )

    galaxy_plotter.figures(contribution_map=True)
    assert path.join(plot_path, "contribution_map.png") in plot_patch.paths


def test__subplots_galaxy_quantities__all_are_output(
    gal_x1_lp_x1_mp,
    sub_grid_7x7,
    grid_irregular_grouped_7x7,
    include_2d_all,
    plot_path,
    plot_patch,
):

    galaxy_plotter = aplt.GalaxyPlotter(
        galaxy=gal_x1_lp_x1_mp,
        grid=sub_grid_7x7,
        include_2d=include_2d_all,
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(plot_path, format="png")),
    )
    galaxy_plotter.subplot_of_light_profiles(image=True)

    assert path.join(plot_path, "subplot_image.png") in plot_patch.paths

    galaxy_plotter.subplot_of_mass_profiles(
        convergence=True, potential=True, deflections_y=True, deflections_x=True
    )

    assert path.join(plot_path, "subplot_convergence.png") in plot_patch.paths
    assert path.join(plot_path, "subplot_potential.png") in plot_patch.paths
    assert path.join(plot_path, "subplot_deflections_y.png") in plot_patch.paths
    assert path.join(plot_path, "subplot_deflections_x.png") in plot_patch.paths
